package edu.gb.mummec.perfectos;

import java.math.BigInteger;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

/**
 *
 * @author german
 */
public class Perfectos {

    static enum Método{NAIF, MEJOR , BUENO, BIG, BIG_THREADED;}
    static Método probar=Método.BIG;    //<---------------------PROBAR
    static final int TOPE_NAIF_MEJOR=100_000; 
    static final int TOPE_BIGS=25_000; 
 
    
    /**
     * @param args the command line arguments
     */
    public static void main(String[] args) {
        long t0,t1;
        t0=System.currentTimeMillis();
        
        switch(probar){
            case NAIF: perfectosNaif(TOPE_NAIF_MEJOR); break;
            case MEJOR: perfectosMejor(TOPE_NAIF_MEJOR); break;
            case BUENO: perfectosBueno(); break;
            case BIG: perfectosBig(TOPE_BIGS); break;
            case BIG_THREADED: perfectosBigThreaded(TOPE_BIGS); break;
        }
        t1=System.currentTimeMillis();
        System.out.println((t1-t0)+" ms.\n----------------");
        
    }
    
    
   /**
     * Calcula perfectos entre 1 y un "tope", probando para cada uno de ellos todos sus divisores. 
     * @param tope número máximo a considerar
     */
    static void perfectosNaif(int tope){
    for (int candidato=1; candidato<tope; candidato++) {
        int suma=0;
        for (int divisor=1;divisor<candidato;divisor++) if (candidato%divisor==0) suma+=divisor;        
        if (candidato==suma) System.out.println(candidato);
    }
}
    
   /**
     * Calcula perfectos entre 1 y un "tope", probando para cada uno los divisores hasta su raiz cuadrada y sumando los encontrados y sus complementarios. 
     * @param tope número máximo a considerar
     */
    static void perfectosMejor(int tope){
    for (int candidato=1; candidato<tope; candidato++) {
        int suma=1;
        for (int divisor=2;divisor<Math.sqrt(candidato);divisor++) if (candidato%divisor==0) suma+=divisor+candidato/divisor;        
        if (candidato==suma) System.out.println(candidato);
    }
}
    
   /**
     * Calcula perfectos mediante su expresión basada en primos de Mersenne [2^(p-1) * (2^p -1)] con (2^p-1) primo hasta la potencia 32. 
     */
    static void perfectosBueno(){
        Intentos:  for (long p=2, terminoA=2, dosP=4; p<=31; terminoA=dosP, dosP<<=1, p++) {
            long terminoB=(dosP-1);
            for (int divisor=2; divisor<Math.sqrt(terminoB);divisor++) if (terminoB%divisor==0) continue Intentos;
            System.out.println(terminoA*terminoB);
        }

}
    
  /**
     * Calcula perfectos mediante su expresión basada en primos de Mersenne [2^(p-1) * (2^p -1)] con (2^p-1) primo, utilizando una representación numérica de precisión indefinida. 
     * OJO. El resultado no es seguro. La primalidad es probabilistica, y se ha fijado a   err&lt;=(1-2^10)
     * Para cada un o presenta: tiempo de ejecución (ms.), p (potencia de mesrenne), número de cifras del perfecto, el perfecto
     * @param maxRo máxima potencia (p en la fórmula)
     */
      
    static void perfectosBig(int maxRo){
        long t0=System.currentTimeMillis();
        int ro;
        BigInteger terminoA, dosPowRo, terminoB;
        String perfecto;
        for (ro=2, terminoA=new BigInteger("2"), dosPowRo=new BigInteger("4"), terminoB=new BigInteger("3");
                 ro<=maxRo;
                 ro++, terminoA=dosPowRo, dosPowRo=dosPowRo.shiftLeft(1), terminoB=(dosPowRo.subtract(BigInteger.ONE)) )
            if (terminoB.isProbablePrime(10)) System.out.printf("%d ms.\t%d\t%d\t%s\n",System.currentTimeMillis()-t0,ro,(perfecto=terminoA.multiply(terminoB).toString()).length(),perfecto);
    }    
    
  /**
     * Calcula perfectos mediante su expresión basada en primos de Mersenne [2^(p-1) * (2^p -1)] con (2^p-1) primo, utilizando una representación numérica de precisión indefinida. 
     * OJO. El resultado no es seguro. La primalidad es probabilistica, y se ha fijado a   err&lt;=(1-2^10)
     * La evaluación de cada candidato se delega en hilos "DeterminaPerfecto" que son ejecutados por un ThreadPool.
     * @param maxRo máxima potencia (p)
     */
    static void perfectosBigThreaded(int maxRo){
        ExecutorService executor = Executors.newFixedThreadPool(7);

        long t0=System.currentTimeMillis();
        int ro;
        BigInteger terminoA, dosPowRo, terminoB;
        String perfecto;
        for (ro=2, terminoA=new BigInteger("2"), dosPowRo=new BigInteger("4"), terminoB=new BigInteger("3");
                 ro<=maxRo;
                 ro++, terminoA=dosPowRo, dosPowRo=dosPowRo.shiftLeft(1), terminoB=(dosPowRo.subtract(BigInteger.ONE)) ) 
            executor.execute(new DeterminaPerfecto(ro, terminoA, dosPowRo, terminoB));
    }    
    
   /**
     * Clase auxiliar que determina e informa si un candidato es perfecto,usando un  hilo. 
     * Para cada un o presenta: tiempo de ejecución (ms.), p (potencia de mesrenne), número de cifras del perfecto, el perfecto
     */
    private static class DeterminaPerfecto implements Runnable {
        int ro;
        long t0=System.currentTimeMillis();
        BigInteger terminoA, dosPowRo, terminoB;

        public DeterminaPerfecto(int ro, BigInteger terminoA, BigInteger dosPowRo, BigInteger terminoB) {
            this.ro=ro;
            this.terminoA = terminoA;
            this.dosPowRo = dosPowRo;
            this.terminoB = terminoB;
        }
        
        @Override
        public void run() {
            String perfecto;
            if (terminoB.isProbablePrime(10)) System.out.printf("%d ms.\t%d\t%d\t%s\n",System.currentTimeMillis()-t0,ro,(perfecto=terminoA.multiply(terminoB).toString()).length(),perfecto);
        }
        
    }
    
    
    
    
    
    
}